-- BOLT - Beacon Of Light Timer
-- by Daltanious of Kul Tiras

BOLT = {}
local BOLT = BOLT

-- used to skip unwanted events
local addonLoadedDone = false

BOLT.events = {}
BOLT.playerName = nil
BOLT.playerClass = nil
BOLT.playerLevel = nil
BOLT.playerInCombat = false

-- ## GLOBAL CONSTANTS ##
BOLT.versionNumber = "1.7"
BOLT.soundFile1 = "Sound\\Interface\\PlayerInviteA.wav" 		-- loud sound alert 
BOLT.soundFile2 = "Sound\\Interface\\PlaceHolder.wav"   		-- soft sound alert
BOLT.soundFile3 = "Sound\\Interface\\Aggro_Pulled_Aggro.wav"	-- dispel sound alert
-- Beacon of Light
BOLT.bolSpellID = 53563
BOLT.bolGlyphSpellID = 63218
BOLT.bolStandardDuration = 299
BOLT.bolGlyphBonus = 30
BOLT.bolSpellName, _, BOLT.bolTexture, _, _, _, _, _, _ = GetSpellInfo( BOLT.bolSpellID )
-- Sacred Shield
BOLT.ssSpellID = 86273
BOLT.ssStandardDuration = 29
BOLT.ssDGRank1Bonus = 15
BOLT.ssDGRank2Bonus = 30
BOLT.ssSpellName, _, BOLT.ssTexture, _, _, _, _, _, _ = GetSpellInfo( BOLT.ssSpellID )
-- Judgements
BOLT.jolSpellID = 20271
BOLT.jowSpellID = 20271
BOLT.jojSpellID = 20271
BOLT.jolSpellName, _, _, _, _, _, _, _, _ = GetSpellInfo( BOLT.jolSpellID )
BOLT.jowSpellName, _, _, _, _, _, _, _, _ = GetSpellInfo( BOLT.jowSpellID )
BOLT.jojSpellName, _, _, _, _, _, _, _, _ = GetSpellInfo( BOLT.jojSpellID )
-- Judgements of the Pure
BOLT.jotpSpellID = 53657
BOLT.jotpStandardDuration = 60
BOLT.jotpSpellName, _, BOLT.jotpTexture, _, _, _, _, _, _ = GetSpellInfo( BOLT.jotpSpellID )
-- slash commands
SLASH_BOLT1 = "/BOLT"
SLASH_BOLT2 = "/BOLTtimer"

-- As general rule: ReadSv( "A.B.C.D" ) will return BOLTsv["A"]["B"]["C"]["D"]
-- MAX 4 nested levels
function ReadSv( svPath )
	local A, B, C, D = strsplit( "." , svPath, 4 )
	
	assert ( ( A ~= nil ) or ( B ~= nil ) or ( C ~= nil ) or ( D ~= nil ) , "ReadSv unknown svPath" )
	
	if ( A ~= nil ) and ( B == nil ) and ( C == nil ) and ( D == nil ) then
		return BOLTsv[A]
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C == nil ) and ( D == nil ) then
		return BOLTsv[A][B]
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C ~= nil ) and ( D == nil ) then
		return BOLTsv[A][B][C]
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C ~= nil ) and ( D ~= nil ) then
		return BOLTsv[A][B][C][D]
	end	
end

function WriteSv( svPath, svValue )
	local A, B, C, D = strsplit( "." , svPath, 4 )
	
	assert ( ( A ~= nil ) or ( B ~= nil ) or ( C ~= nil ) or ( D ~= nil ) , "WriteSv unknown svPath" )
	
	if ( A ~= nil ) and ( B == nil ) and ( C == nil ) and ( D == nil ) then
		BOLTsv[A] = svValue
		return
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C == nil ) and ( D == nil ) then
		BOLTsv[A][B] = svValue
		return
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C ~= nil ) and ( D == nil ) then
		BOLTsv[A][B][C] = svValue
		return
	end
	
	if ( A ~= nil ) and ( B ~= nil ) and ( C ~= nil ) and ( D ~= nil ) then
		BOLTsv[A][B][C][D] = svValue
		return
	end	
end

function ToggleBoolSv( svPath )
	local value
	
	value = not ReadSv( svPath )
	WriteSv( svPath , value  )
	return value
end

-- Check BOLT.events[] for the function
BOLT.eventFrame = CreateFrame( "Frame" )
BOLT.eventFrame:SetScript( "OnEvent", function( this, event, ... )
  BOLT.events[event]( ... )
end )

-- Main event
BOLT.eventFrame:RegisterEvent( "ADDON_LOADED" )
BOLT.eventFrame:RegisterEvent( "VARIABLES_LOADED" )
BOLT.eventFrame:RegisterEvent( "PLAYER_ALIVE" )

-- ### EVENTS HANDLERS ###
function BOLT.events.ADDON_LOADED( addon )
	local abort = false
		
	if addon ~= "BOLT" then return end
	
	print( "BOLT v" ..BOLT.versionNumber .. BOLT_LOADED )
	-- BOLT.playerClass = UnitClass( "player" )
	-- BOLT.playerLevel = UnitLevel( "player" )
	-- if BOLT.playerClass ~= "Paladin" then	
	BOLT.playerName = UnitName( "player" )
	_, BOLT.playerClass = UnitClass( "player" )
	BOLT.playerLevel = UnitLevel( "player" )
	if BOLT.playerClass ~= "PALADIN" then	
		print( BOLT_NO_PALADIN )
		abort = true
	else
		if BOLT.playerLevel < 60 	then
			-- 60 is the minimum level for Beacon of Light
			-- 80 is the minimum level for Sacred Shield
			print( BOLT_NO_60 )		
			abort = true
		end
	end
	-- abort the loading process
	if abort then
		BOLT.eventFrame:UnregisterAllEvents()
		print( "BOLT v" ..BOLT.versionNumber .. BOLT_UNLOADED )
		return
	end
	
	-- saved variables container initialization
	if BOLTsv == nil then
		BOLTsv = {}
	end	
	if ReadSv( "generalOptions" ) == nil then
		WriteSv( "generalOptions" , {} )
	end
	if ReadSv( "generalOptions.soundFile" ) == nil then
		WriteSv( "generalOptions.soundFile" , BOLT.soundFile1 )			-- default sound file
	end
	if ReadSv( "generalOptions.frameStrata" ) == nil then
		WriteSv( "generalOptions.frameStrata" , "BACKGROUND" )			-- default frame strata
	end
	if ReadSv( "generalOptions.dimAlpha" ) == nil then
		WriteSv( "generalOptions.dimAlpha" , 0.5 )						-- default alpha value for dimmed buttons 
	end
	if ReadSv( "generalOptions.hideWhenOOC" ) == nil then
		WriteSv( "generalOptions.hideWhenOOC" , false )					-- default no auto hide when out of combat
	end
	if ReadSv( "generalOptions.alwaysShowNamePlates" ) == nil then
		WriteSv( "generalOptions.alwaysShowNamePlates" , false )		-- default show name plates only during a countdown
	end
	if ReadSv( "generalOptions.removeRoundCorners" ) == nil then
		WriteSv( "generalOptions.removeRoundCorners" , false )			-- default do not remove round corners
	end	

	-- other events watched
	BOLT.eventFrame:RegisterEvent( "UNIT_SPELLCAST_SENT" )
	BOLT.eventFrame:RegisterEvent( "PLAYER_TALENT_UPDATE" )
	BOLT.eventFrame:RegisterEvent( "PLAYER_REGEN_DISABLED" )
	BOLT.eventFrame:RegisterEvent( "PLAYER_REGEN_ENABLED" )
	
	-- registering slash commands
	SlashCmdList["BOLT"] = BOLT.Options
	addonLoadedDone = true
end

-- when wow is loaded for the first time time this is the first event where talent informations
-- are available ( readable ), unfortunately is fired only once during the first loading process ( cold start )
-- and when other events occours like char resurrect and so on... 
-- here I don't want to create BOLTBUttons on events different from a simple cold start
do
	local onlyOnce = true
	function BOLT.events.PLAYER_ALIVE( addon )
		if onlyOnce then
			BOLT.BOL = BOLTButton:Create( "BOL", 0 )
			BOLT.SS = BOLTButton:Create( "SS", 1 )
			BOLT.JOTP = BOLTButton:Create( "JOTP", 2 )
			-- option frame
			BOLT.CreateOptionFrame()
			-- set visibility accordingly the hide when out of combat flag
			BOLT.ShowHide()
			onlyOnce = false
		end
	end	
	function Skip_PLAYER_ALIVE()
		onlyOnce=false
	end
end

-- return true/false if talent information are available or not
local function AreTalentInfosAvailable()
		local n, rank
		
		n = 0
		for j= 1, 3 do
			for i= 1, 26 do
				_, _, _, _, rank, _= GetTalentInfo( j, i )
				n = n + rank
			end
		end
		return n ~= 0
end

-- VARIABLES_LOADED is fired on cold start... but the talent informations are not available ( that's
-- why I use PLAYER_ALIVE to handle cold starts ). On the other side this event is fired every reloadUI
-- with talent informations available... so here I skip BOLTButton creation on cold start only
function BOLT.events.VARIABLES_LOADED( addon )	
	if AreTalentInfosAvailable() then
		-- if talent informations are available the VARIABLES_LOADED is fired through a reloadui event
		BOLT.BOL = BOLTButton:Create( "BOL", 0 )
		BOLT.SS = BOLTButton:Create( "SS", 1 )
		BOLT.JOTP = BOLTButton:Create( "JOTP", 2 )
		-- option frame
		BOLT.CreateOptionFrame()
		-- set visibility accordingly the hide when out of combat flag
		BOLT.ShowHide()
		-- reloadui reset the local "static" variables... now here i must force
		-- to skip the PlayerAlive
		Skip_PLAYER_ALIVE()
	end
end

-- again another event to handle carefully... PLAYER_TALENT_UPDATE occours on talent change AND
-- immediately after the ADDON_LOADED event... in this second case I don't want to manage button visibility
function BOLT.events.PLAYER_TALENT_UPDATE( tabID, pos )
	if addonLoadedDone==true then
		BOLT.BOL:ApplyVisibilityTemplate()
		BOLT.SS:ApplyVisibilityTemplate()
		BOLT.JOTP:ApplyVisibilityTemplate()
	end
	addonLoadedDone = false
end

function BOLT.events.PLAYER_REGEN_DISABLED()	
	-- restore the default button alpha value
	BOLT.DimAll()
	BOLT.playerInCombat = true
end

function BOLT.events.PLAYER_REGEN_ENABLED()
	-- restore the default button alpha value
	if ReadSv( "generalOptions.hideWhenOOC" ) then
		BOLT.HideAll()		
	end
	BOLT.playerInCombat = false
end

function BOLT.DimAll()
	BOLT.BOL:Dim()
	BOLT.SS:Dim()
	BOLT.JOTP:Dim()
end

function BOLT.HideAll()
	BOLT.BOL:Hide()
	BOLT.SS:Hide()
	BOLT.JOTP:Hide()
end

function BOLT.RemoveAllRoundCorners()
	BOLT.BOL:RemoveRoundCorners()
	BOLT.SS:RemoveRoundCorners()
	BOLT.JOTP:RemoveRoundCorners()
end

function BOLT.SetAllRoundCorners()
	BOLT.BOL:SetRoundCorners()
	BOLT.SS:SetRoundCorners()
	BOLT.JOTP:SetRoundCorners()
end

-- dispatcher
function BOLT.events.UNIT_SPELLCAST_SENT( unitID, spell, rank, target )
	local targetClass
	
	if unitID ~= "player"  then return end
	
	if spell == BOLT.bolSpellName then	
		-- Beacon of Light successfully casted.. start the duration timer
		BOLT.BOL:StartTimer( target )
		return
	end
	
	if spell == BOLT.ssSpellName then	
		-- Sacred Shield successfully casted.. start the duration timer
		BOLT.SS:StartTimer( target )
		return
	end	
	
	if ( spell == BOLT.jolSpellName ) or ( spell == BOLT.jowSpellName ) or ( spell == BOLT.jojSpellName )then
		BOLT.JOTP:StartTimer( target )
		return
	end
end

function BOLT.SetDimAlphaValue( newAlpha )
	assert( ( newAlpha>=0 ) and ( newAlpha<=1 ), "BOLT.SetDimAlphaValue bad alpha value!" )
	
	WriteSv( "generalOptions.dimAlpha" , newAlpha )
	BOLT.DimAll()
end

function BOLT.ShowHide()
	if not BOLT.playerInCombat and ReadSv( "generalOptions.hideWhenOOC" ) then
		BOLT.HideAll()
	else
		BOLT.DimAll()	
	end
end

function BOLT.ToggleAutoHideOOC()
	ToggleBoolSv( "generalOptions.hideWhenOOC" )
	BOLT.ShowHide()
end

function BOLT.ToggleAlwaysShowNamePlates()
	ToggleBoolSv( "generalOptions.alwaysShowNamePlates" )
	if ReadSv( "generalOptions.alwaysShowNamePlates" ) then
		BOLT.BOL:ShowMacroTarget()
		BOLT.SS:ShowMacroTarget()
	else
		BOLT.ShowHide()
	end
end

function BOLT.ToggleRoundCorners()
	ToggleBoolSv( "generalOptions.removeRoundCorners" )
	if ReadSv( "generalOptions.removeRoundCorners" ) then
		BOLT.RemoveAllRoundCorners()
	else
		BOLT.SetAllRoundCorners()
	end
end

do
	-- "static" declarations
	local c	= 1 	-- counter for unique names	
	local currLine = { d1=0, d2=0, d3=0, d4=0, d5=0 }	-- current available line for option screens 1..5
	
	local function GetPage( pageNum )
		assert( ( pageNum >= 1 ) and ( pageNum <= 5 ), "GetPage bad pageNum" )
		
		return currLine["d" .. pageNum]		
	end
	
	local function SetPage( pageNum, value )
		assert( ( pageNum >= 1 ) and ( pageNum <= 5 ), "GetPage bad pageNum" )
		
		value = value - 30	-- set the value for the next line
		c = c + 1
		currLine["d" .. pageNum] = value	
	end
  
	local function Round( v )
		return floor( v * 10 + 0.5 ) / 10
	end
  	
	function Text( pageNum, panel, text, font, cr )
		local d, fs
  	
		d = GetPage( pageNum )	
		fs = panel:CreateFontString( "BOLT_fs" .. c, "OVERLAY", font )
		fs:SetText( text )
		fs:SetPoint( "TOPLEFT", 10, d - 20 )
		if cr then	-- carriage return
			SetPage( pageNum, d )
		end		
		return d
	end
	
	function TextAndCheck( pageNum, panel, text, ToggleFunction, initialStatus )
		local d, fs, cb
		
		d = Text ( pageNum, panel, text, "GameFontNormal", false)		
		cb = CreateFrame( "CheckButton", "$parent_cb" .. c, panel, "OptionsCheckButtonTemplate" )
		cb:SetWidth( 18 )
		cb:SetHeight( 18 )
		cb:SetScript( "OnClick", ToggleFunction )
		cb:SetPoint( "TOPRIGHT", -10, d - 20 )
		cb:SetChecked( initialStatus )
		SetPage( pageNum, d )
	end	
	
	function TextAndSlide( pageNum, panel, text, ToggleFunction, initialValue, minV, maxV, stepV )
		local d, fs, sl
		
		d = Text ( pageNum, panel, text, "GameFontNormal", false)		
		sl = CreateFrame( "Slider", "$parent_sl" .. c, panel, "OptionsSliderTemplate" )
		sl:SetMinMaxValues( minV, maxV )
		sl:SetValue( initialValue )
		sl:SetValueStep( stepV )
		sl:SetScript( "OnValueChanged", function( self ) ToggleFunction( Round( self:GetValue() ) ); getglobal( self:GetName() .. "Text" ):SetText( Round( self:GetValue() ) )  end )
		getglobal( sl:GetName() .. "Low" ):SetText( minV )
		getglobal( sl:GetName() .. "High" ):SetText( maxV )
		getglobal( sl:GetName() .. "Text" ):SetText( initialValue )
		sl:SetPoint( "TOPRIGHT", -10, d - 20 )
		SetPage( pageNum, d )	
	end	
	
end

function BOLT.CreateOptionFrame()
	local panel = CreateFrame( "FRAME", "BOLTOptions" )
	panel.name = "BOLT"

	local panel2 = CreateFrame( "FRAME", "BOLTGlobalOptions" )
	panel2.name = BOLT_NAME_PANEL_2
	panel2.parent = "BOLT"

	local panel3 = CreateFrame( "FRAME", "BOLTTimersOptions" )
	panel3.name = BOLT_NAME_PANEL_3
	panel3.parent = "BOLT"
	
	local panel5 = CreateFrame( "FRAME", "BOLTTimersOptions" )
	panel5.name = BOLT_NAME_PANEL_5
	panel5.parent = "BOLT"

	-- PANEL #1
	Text( 1, panel, "BOLT v" .. BOLT.versionNumber .. " - by Daltanious of Kul Tiras", "GameFontRedLarge", true )	  
	Text( 1, panel, "Rereleased by Illisis of Barthilas", "GameFontRedLarge", true )	
	Text( 1, panel, "Credits:", "GameFontNormal", true )
	Text( 1, panel, "Thaeryn of Dethecus & Tekkub for the code examples", "GameFontNormal", true )
	Text( 1, panel, "OlegKR for the internationalization code and russian translation", "GameFontNormal", true )
	Text( 1, panel, "Controls:", "GameFontRedLarge", true )
	Text( 1, panel, "SHIFT + Doubleclick to Pin/Unpin a Button on the screen", "GameFontNormal", true )
	Text( 1, panel, "CTRL + Doubleclick to set up a target as focus for BOL", "GameFontNormal", true )

	-- PANEL #2
	-- Alpha dim value
	TextAndSlide( 2, panel2, BOLT_ALPHA_LEVEL, function( scaleParameter ) BOLT.SetDimAlphaValue( scaleParameter ) end, ReadSv( "generalOptions.dimAlpha" ), 0.2, 0.8, 0.1 )
	-- Hide button when out of combat
	TextAndCheck( 2, panel2, BOLT_HIDE_BUTTONS, function() BOLT.ToggleAutoHideOOC() end, ReadSv( "generalOptions.hideWhenOOC" ) )
	-- Always show name plates
	TextAndCheck( 2, panel2, BOLT_SHOW_NAME, function() BOLT.ToggleAlwaysShowNamePlates() end, ReadSv( "generalOptions.alwaysShowNamePlates" ) )
	-- Remove button round corners
	TextAndCheck( 2, panel2, BOLT_BUTTONS_CORNERS, function() BOLT.ToggleRoundCorners() end, ReadSv( "generalOptions.removeRoundCorners" ) )

	-- PANEL #3     
	-- Beacon of Light sound alert  
	TextAndCheck( 3, panel3, BOLT_BL_SOUND_ALERT , function() BOLT.BOL:ToggleSoundAlert() end, BOLT.BOL:GetSoundOption() )
	-- Beacon of Light flash alert
	TextAndCheck( 3, panel3, BOLT_BL_FLASH_ALERT, function() BOLT.BOL:ToggleFlashAlert() end, BOLT.BOL:GetFlashOption() )
	-- Beacon of Light visiblity
	TextAndCheck( 3, panel3, BOLT_BL_HOLY_SPEC, function() BOLT.BOL:ToggleVisibility( "h" ) end, BOLT.BOL:GetVisibility( "h" ) )
	TextAndCheck( 3, panel3, BOLT_BL_PROTECT_SPEC, function() BOLT.BOL:ToggleVisibility( "p" ) end, BOLT.BOL:GetVisibility( "p" ) )
	TextAndCheck( 3, panel3, BOLT_BL_RETRIBUT_SPEC, function() BOLT.BOL:ToggleVisibility( "r" ) end, BOLT.BOL:GetVisibility( "r" ) )  
	-- Beacon of Light scale
	TextAndSlide( 3, panel3, BOLT_BL_BUTTON, function( scaleParameter ) BOLT.BOL:SetFrameScale( scaleParameter ) end, BOLT.BOL:GetFrameScale(), 0.3, 1.8, 0.1 )	 
	
	-- PANEL #5
	-- Judgements of the Pure sound alert
	TextAndCheck( 5, panel5, BOLT_JP_SOUND_ALERT, function() BOLT.JOTP:ToggleSoundAlert() end, BOLT.JOTP:GetSoundOption() )
	-- Judgements of the Pure flash alert
	TextAndCheck( 5, panel5, BOLT_JP_FLASH_ALERT, function() BOLT.JOTP:ToggleFlashAlert() end, BOLT.JOTP:GetFlashOption() )		
	-- Judgements of the Pure visiblity
	TextAndCheck( 5, panel5, BOLT_JP_HOLY_SPEC, function() BOLT.JOTP:ToggleVisibility( "h" ) end, BOLT.JOTP:GetVisibility( "h" ) )
	TextAndCheck( 5, panel5, BOLT_JP_PROTECT_SPEC, function() BOLT.JOTP:ToggleVisibility( "p" ) end, BOLT.JOTP:GetVisibility( "p" ) )
	TextAndCheck( 5, panel5, BOLT_JP_RETRIBUT_SPEC, function() BOLT.JOTP:ToggleVisibility( "r" ) end, BOLT.JOTP:GetVisibility( "r" ) )  			    
	-- Judgements of the Pure scale
	TextAndSlide( 5, panel5, BOLT_JP_BUTTON, function( scaleParameter ) BOLT.JOTP:SetFrameScale( scaleParameter ) end, BOLT.JOTP:GetFrameScale(), 0.3, 1.8, 0.1 )    
		
	-- add panels to the interface option screen
	InterfaceOptions_AddCategory(panel)
	InterfaceOptions_AddCategory(panel2)
	InterfaceOptions_AddCategory(panel3)
	InterfaceOptions_AddCategory(panel5)
end





-- ### SLASH COMMAND HANDLER ###
function BOLT.Options( msg, editbox )
	if msg==nil then msg = ""	end
	msg = string.lower( msg ) 
	if ( msg=="" ) or ( msg=="help" ) then
		print( "BOLT v" .. BOLT.versionNumber .. " - by Daltanious of Kul Tiras - Rereleased by Illisis of Barthilas" )
		InterfaceOptionsFrame_OpenToCategory( "BOLT" )
		return
	end
end