-- BOLT BUTTON CLASS IMPLEMENTATION

BOLTButton = {}
BOLTButton.__index = BOLTButton
local BOLTButton = BOLTButton

local MACRO_HEAD = "/cast [target="
local MACRO_BOL_TAIL = "] " .. BOLT.bolSpellName
local MACRO_SS_TAIL = "] " .. BOLT.ssSpellName

-- CONSTRUCTOR
function BOLTButton:Create( btnName, btnOrder )
	local self = {}
	local frameBtn
	local displayFont = "Fonts\\FRIZQT__.TTF"	-- display font		
		
	setmetatable( self, BOLTButton )			
	-- just a quick check of parameters
	assert( ( btnName == "BOL" ) or ( btnName == "SS" ) or ( btnName == "JOTP" ), "BOLTButton:Create unknown button name!" )
	assert( ( btnOrder >= 0 ) and ( btnOrder  <= 2 ), "BOLTButton:Create bad button order!" )	
	
	-- CLASS NAME
	self.name = btnName
	
	-- DISPLAY BUTTON
	frameBtn = CreateFrame( "Button", nil, UIParent, "SecureActionButtonTemplate" )	
	frameBtn:SetFrameStrata( ReadSv("generalOptions.frameStrata" ) )
	
	-- RIGHT CLICK on the frame will launch the macro
	frameBtn:SetAttribute( "type2", "macro" )
	--frameBtn:SetAttribute( "type", "macro" )
	-- enable the button to receive any of these clicks:
	frameBtn:RegisterForClicks( "LeftButtonUp", "RightButtonUp", "LeftButtonDown", "RightButtonDown" )	
	if btnName == "BOL" then
		self.macroTail = MACRO_BOL_TAIL
	else
		if btnName == "SS" then 
			self.macroTail = MACRO_SS_TAIL
		else
			-- JOTP
			self.macroTail = ""	-- no refresh ability for JOTP
		end
	end
		
	frameBtn:SetWidth( 48 )
	frameBtn:SetHeight( 48 )
	frameBtn:EnableMouse( true )	
	frameBtn:SetClampedToScreen( true )
	frameBtn.texture = frameBtn:CreateTexture( nil, "ARTWORK" )		
	if btnName == "BOL" then
		frameBtn.texture:SetTexture( BOLT.bolTexture )
	else
		if btnName == "SS" then
			frameBtn.texture:SetTexture( BOLT.ssTexture )
		else
			-- JOTP
			frameBtn.texture:SetTexture( BOLT.jotpTexture )
		end
	end	
	frameBtn.texture:ClearAllPoints()
	frameBtn.texture:SetAllPoints( frameBtn )
	frameBtn.texture:SetAlpha( ReadSv( "generalOptions.dimAlpha" ) )  -- dimmed by deafult
	frameBtn.durationText = frameBtn:CreateFontString( nil, "OVERLAY", nil )
	frameBtn.durationText:SetJustifyH( "CENTER" )

	if ( self.name == "BOL" ) or ( self.name == "SS" ) then
		frameBtn.durationText:SetPoint( "CENTER", 1, 5 )	
	else
		frameBtn.durationText:SetPoint( "CENTER", 1, 0 )	
	end

	frameBtn.durationText:SetFont( displayFont, 32, "THICKOUTLINE" )	
	frameBtn.durationText:SetTextColor( 1, 1, 1, 1 )
	frameBtn.durationText:SetText( "" ) 							-- no text by default	
	frameBtn.namePlate = frameBtn:CreateFontString( nil, "OVERLAY", nil )
	frameBtn.namePlate:SetJustifyH( "CENTER" )
	frameBtn.namePlate:SetPoint( "CENTER", 1, -14 )
	frameBtn.namePlate:SetFont( displayFont, 13, "OUTLINE" )
	frameBtn.namePlate:SetTextColor( 1, 1, 1, 1 )
	frameBtn.namePlate:SetText( "" ) 								-- no text by default
	-- general button options
	if self:GetSv( "options" ) == nil then
		self:SetSv( "options", {} )
	end
	if self:GetSv( "options.movable" ) == nil then
		self:SetSv( "options.movable", true )						-- unlocked by default
	end
	if self:GetSv( "options.scale" ) == nil then
		self:SetSv( "options.scale", 1.0 )							-- deafult scale
	end
		
	-- set the initial position or restore it from saved variables
	frameBtn:ClearAllPoints()
	if self:GetSv( "framePos" ) == nil then
		frameBtn:SetPoint( "CENTER", 100 * ( btnOrder - 1 ), 0 )
		self:SetSv( "framePos", {cX = frameBtn:GetLeft() * self:GetSv( "options.scale" ), 
								 cY = frameBtn:GetBottom() * self:GetSv( "options.scale" )}	)	
	else
		frameBtn:SetScale( self:GetSv( "options.scale" ) )
		frameBtn:SetPoint( "BOTTOMLEFT", self:GetSv( "framePos.cX" ) / self:GetSv( "options.scale" ), 
										 self:GetSv( "framePos.cY" ) / self:GetSv( "options.scale" ) )
	end
						
	-- custom default values for BOL & SS buttons	
	if self:GetSv( "visibility" ) == nil then
		if ( btnName == "BOL" ) or ( btnName == "JOTP" ) then 
			self:SetSv( "visibility" , {h=true, p=false, r=false} )
		else
			-- SS
			self:SetSv( "visibility" , {h=false, p=false, r=false} )
		end
	end
	if self:GetSv( "alerts" ) == nil then
		self:SetSv( "alerts" , {} )
	end	
	if btnName == "BOL" then 
		if self:GetSv( "alerts.sound" ) == nil then
			self:SetSv( "alerts.sound" , true )
		end
		if self:GetSv( "alerts.flash" ) == nil then
			self:SetSv( "alerts.flash" , true )
		end
	else
		-- SS & JOTP
		if self:GetSv( "alerts.sound" ) == nil then
			self:SetSv( "alerts.sound", false )
		end
		if self:GetSv( "alerts.flash" ) == nil then
			self:SetSv( "alerts.flash", true )
		end	
	end		
	if self:GetSv( "macroTarget" ) == nil then
		self:SetSv( "macroTarget", "" )
	end
		
	-- FRAME BUTTON MOVEMENT HANDLERS
	if  self:GetSv( "options.movable" ) then
		frameBtn:SetMovable( true )
		frameBtn:SetScript( "OnMouseDown", function( self ) self:StartMoving() end )		
		frameBtn:SetScript( "OnMouseUp", function() BOLTButton.OnLeftClick( self ) end )
	else	
		frameBtn:SetMovable( false )
		frameBtn:SetScript( "OnMouseDown", nil )
		frameBtn:SetScript( "OnMouseUp", nil )
	end
	-- double click on the frame will switch the movable status
	frameBtn:SetScript( "OnDoubleClick", function( _, btn ) BOLTButton.OnDoubleClick( self, btn ) end )

	-- by default, every button is not visible, check ApplyVisibilityTemplate
	self.visible = false
	
	-- bind the temp button to the class
 	self.frameBtn = frameBtn		

	-- COOLDOWN frame used only for the flashing animation	
	self.frameCd = CreateFrame( "Cooldown", nil, frameBtn )
	self.frameCd:SetAllPoints( frameBtn )
	self.frameCd:SetFrameStrata( "BACKGROUND" )		
	-- TIMER
	self.timerFrame = CreateFrame( "Frame" )
	-- other timer parameters
	self.targetName = nil
	self.targetClass = nil
	self.duration = nil
	self.currentDuration = nil
	self.timeCounter = nil	
	-- MACRO	
	self:SetNewSpellTarget(nil)	
	-- VISIBILITY
	self:ApplyVisibilityTemplate()	
	-- GRAY NAME PLATES
	if ReadSv( "generalOptions.alwaysShowNamePlates" ) then
		self:ShowMacroTarget()
	end	
	-- set or remove round corners from button textures
	if ReadSv( "generalOptions.removeRoundCorners" ) then
		self:RemoveRoundCorners()
	else
		self:SetRoundCorners()
	end
	
 	return self
end

-- called outside the object to start a duration count
function BOLTButton:StartTimer( targetName )
	self.targetName = targetName
	_, self.targetClass = UnitClass( targetName )
	self.duration = self:GetSpellDuration()
	self.currentDuration = self.duration
	self.timeCounter = 0
	self.timerFrame:SetScript( "OnUpdate", nil )
	self.timerFrame:SetScript( "OnUpdate", function( _, elapsed ) BOLTButton.Timer( self, elapsed ) end )
--	self.frameCd:SetCooldown( GetTime(), self.duration )	
end

-- tic tac...
function BOLTButton.Timer( self, elapsed )
	self.timeCounter = self.timeCounter + elapsed
	if self.timeCounter >= 1 then
		self.currentDuration = self.currentDuration -1
		self.timeCounter = 0
		self:Synchronize()
	end
	if self.currentDuration <= 0 then
		self:AbortTimer()
	end
end

function BOLTButton:AbortTimer()
	self.currentDuration = self.duration
	self.currentDuration = 0		
	self.timerFrame:SetScript( "OnUpdate", nil )
end

-- syncronize the button with his curent status
function BOLTButton:Synchronize()
	local buff

	if not self.visible then
		return
	else
		-- first of all we check if the buff still exist
		if self.name == "BOL" then
			buff ,_ ,_ ,_ ,_ ,_ ,_ ,_ ,_ = UnitBuff( self.targetName, BOLT.bolSpellName )
		else
			if self.name == "SS" then
				buff ,_ ,_ ,_ ,_ ,_ ,_ ,_ ,_ = UnitBuff( self.targetName, BOLT.ssSpellName )		
			else
				-- JOTP
				buff ,_ ,_ ,_ ,_ ,_ ,_ ,_ ,_ = UnitBuff( UnitName( "player" ) , BOLT.jotpSpellName )				
			end
		end
		
		-- and ( self.currentDuration < self.duration - 1 ) => added to solve lag problems
		if ( buff == nil ) and ( self.currentDuration > 0 ) and ( self.currentDuration < self.duration - 1 ) then
			-- the buff was dispelled / removed
			self:AbortTimer()			
			PlaySoundFile( BOLT.soundFile3 )
			self.frameCd:SetCooldown( 0, 0 )			
		end
		
		if ( self.currentDuration <= 10 ) and ( buff ~= nil ) then
			-- dum dum dum...
			if self:GetSv( "alerts.sound" ) then
				PlaySoundFile( ReadSv( "generalOptions.soundFile" ) )	
			end
			-- flash alert
			if self:GetSv( "alerts.flash" ) then
				self.frameCd:SetCooldown( 0, 0 )				
			end
		end
		
		if self.currentDuration > 0 then
			-- I'm inside a duration count			
			self:SetCounter()
			if self.currentDuration + 1 == self.duration then
				-- just launched the timer...
				self:SetNamePlate()
			end			
		else
			-- I'm not inside a duration count
			self:Reset()
			if not BOLT.playerInCombat and ReadSv( "generalOptions.hideWhenOOC" ) then
				self:Hide()
			else
				self:Dim()
			end
		end
	end		
end

-- left click handler
function BOLTButton:OnLeftClick()
	self.frameBtn:StopMovingOrSizing()
	self:SetSv( "framePos", {cX = self.frameBtn:GetLeft() * self:GetSv( "options.scale" ), 
							 cY = self.frameBtn:GetBottom() * self:GetSv( "options.scale" )} )
	self:SetFrameScale(nil)
end

-- double clicks handler
-- CTRL + DoubleClick => set new macro target (only for BOL & SS)
-- SHIFT + DoubleClick => lock/unlock frame position
function BOLTButton.OnDoubleClick(self, btn)
	local v
	-- discard every double click different from a leftbutton double click	
	if btn ~= "LeftButton" then	return end
	
	-- if IsAltKeyDown() and ( self.currentDuration > 0 ) then
		-- self.frameBtn.namePlate:SetAlpha( 0 )		-- hide the name plate
		-- self.frameBtn.namePlate:SetText( "" )		-- reset the name plate
		-- return
	-- end
		
	if IsShiftKeyDown() then
		if self:ToggleBoolSv( "options.movable" ) == false then
			self.frameBtn:SetMovable( false )
			self.frameBtn:SetScript( "OnMouseDown", nil )
			self.frameBtn:SetScript( "OnMouseUp", nil )		
			print( BOLT_FRAME .. self.name .. BOLT_FRAME_LOCKED )
		else
			self.frameBtn:SetMovable( true )
			self.frameBtn:SetScript( "OnMouseDown", function( self ) self:StartMoving() end )
			self.frameBtn:SetScript( "OnMouseUp", function() BOLTButton.OnLeftClick( self ) end )
			print( BOLT_FRAME .. self.name .. BOLT_FRAME_UNLOCKED )
		end
		return
	end
	
	-- JOTP have no target
	if self.name == "JOTP" then return end
	
	if IsControlKeyDown() then	
		if UnitName( "playertarget" ) == nil then
			self:SetNewSpellTarget( UnitName( "player" ) )
		else
			self:SetNewSpellTarget( UnitName( "playertarget" ) )
		end	
		if ReadSv( "generalOptions.alwaysShowNamePlates" ) then
			self:ShowMacroTarget()
		end
		print( BOLT_NEW .. self.name .. BOLT_NEW_TARGET .. self:GetSv( "macroTarget" ) )		
		return
	end		
end

-- ### SPELL RELATED FUNCTIONS / METHODS ###

function BOLTButton:IsBolAvailable()
	local i = 1
	local spellName
	
	while true do
		spellName, _ = GetSpellName( i, BOOKTYPE_SPELL )
		if not spellName then
			return false
		else
			if spellName == BOLT.bolSpellName then
				return true
			end
		end
		i = i + 1
	end
end

function BOLTButton:IsSsAvailable()
	return false
end

function BOLTButton:IsJotpAvailable()
	return true
end

function BOLTButton:GetSpellDuration()
	local enabled, glyphSpellID, link
	local currRank
	
	assert( ( self.name == "BOL" ) or ( self.name == "SS" ) or ( self.name == "JOTP" ), "GetSpellDuration unknown spell!" )
	
	-- Beacon of Light
	if self.name == "BOL" then
		return BOLT.bolStandardDuration
	end
	
	-- Sacred Shield
	if self.name == "SS" then
		return BOLT.ssStandardDuration
	end
	
	if self.name == "JOTP" then
		return BOLT.jotpStandardDuration
	end
end

-- save the current target as new target, only BOL & SS have a proper target
function BOLTButton:SetNewSpellTarget( newMacroTarget )
	-- it work only for BOL & SS
	if ( self.name == "BOL" ) or ( self.name == "SS" ) then
		if newMacroTarget ~= nil then
			self:SetSv( "macroTarget", newMacroTarget )
		end
		self.macroText = MACRO_HEAD .. self:GetSv( "macroTarget" ) .. self.macroTail
		self.frameBtn:SetAttribute( "macrotext", self.macroText )
	end
end

-- this function return 
-- "h" for holy paladins
-- "p" for protection paladins
-- "r" for retribution
-- "u" for unknown
-- Each talent tree have 26 different talents, 71 talent points max
local function GetPalaSpec()
	local rank
	local holy = 0
	local prot = 0
	local retri = 0
	
	for i= 1, 26 do
		_, _, _, _, rank, _= GetTalentInfo( 1, i )
		holy = holy + rank
		_, _, _, _, rank, _= GetTalentInfo( 2, i )
		prot = prot + rank
		_, _, _, _, rank, _= GetTalentInfo( 3, i )
		retri = retri + rank			
	end
	
	if ( holy >= prot ) and ( holy >= retri ) then
		return "h" 	-- holy pala
	end
	if ( prot >= holy ) and ( prot >= retri )	then	
		return "p" 	-- protection pala
	end
	if ( retri >= holy ) and ( retri >= prot ) then		
		return "r" 	-- retribution pala
	end
	return "u" 		-- unknown spec
end

function BOLTButton:ApplyVisibilityTemplate()
	local spec, vis

	-- here we override the visibility template id the related spell/talent is not available

	
	spec = GetPalaSpec()
	if spec == "u" then
		-- button visible if spec is unknown
		vis = true
	else
		vis = self:GetSv( "visibility." .. spec )
	end
	if vis then
		self:Enable()
	else
		self:Disable()
	end		
end

-- spec = H-P-R
-- return true if the button is visible with the passed spec
function BOLTButton:GetVisibility( spec )
	assert( ( spec == "h" ) or ( spec == "p" ) or ( spec == "r" ), "BOLTButton:GetVisibility bad spec value!" )
	return self:GetSv( "visibility." .. spec )
end

function BOLTButton:ToggleVisibility( spec )
	assert( ( spec == "h" ) or ( spec == "p" ) or ( spec == "r" ), "BOLTButton:ToggleVisibility bad spec value!" )	
	self:ToggleBoolSv( "visibility." .. spec )
	self:ApplyVisibilityTemplate()	
end

-- SIMPLE UI MANAGEMENT METHODS

function BOLTButton:Enable()
	self.frameBtn:Show()
	self.visible = true 	-- current visibility status
end

function BOLTButton:Disable()
	self.frameBtn:Hide()
	self.visible = false	-- current visibility status		
end

function BOLTButton:GetFrameScale()
	return self:GetSv( "options.scale" )
end

function BOLTButton:SetFrameScale( newScale )
	assert( ( newScale == nil ) or ( newScale >= 0.1 ) and ( newScale <= 2 ), "BOLTButton:SetFrameScale bad scale value" )
	
	if newScale ~= nil then
		self:SetSv( "options.scale", newScale )
	end
	self.frameBtn:SetScale( self:GetSv( "options.scale" ) )
	self.frameBtn:ClearAllPoints()
	self.frameBtn:SetPoint( "BOTTOMLEFT", self:GetSv( "framePos.cX" ) / self:GetSv( "options.scale" ), self:GetSv( "framePos.cY" ) / self:GetSv( "options.scale" ) )
end

function BOLTButton:GetSoundOption()
	return self:GetSv( "alerts.sound" )
end

function BOLTButton:ToggleSoundAlert()	
	if self:ToggleBoolSv( "alerts.sound" ) then
		PlaySoundFile( ReadSv( "generalOptions.soundFile" ) )
	end	
end

function BOLTButton:GetFlashOption()
	return self:GetSv( "alerts.flash" )
end

function BOLTButton:ToggleFlashAlert()	
	if self:ToggleBoolSv( "alerts.flash" ) then
		self.frameCd:SetCooldown( 0, 0 )		
	end	
end

-- This function can return a substring of a UTF-8 string, properly handling
-- UTF-8 codepoints.  Rather than taking a start index and optionally an end
-- index, it takes the string, the start index and the number of characters
-- to select from the string.
--
-- UTF-8 Reference: 
-- 0xxxxxxx - 1 byte UTF-8 codepoint (ASCII character)
-- 110yyyxx - First byte of a 2 byte UTF-8 codepoint
-- 1110yyyy - First byte of a 3 byte UTF-8 codepoint
-- 11110zzz - First byte of a 4 byte UTF-8 codepoint
-- 10xxxxxx - Inner byte of a multi-byte UTF-8 codepoint
local function utf8sub(str, start, numChars)
  local currentIndex = start
  while numChars > 0 and currentIndex <= #str do
    local char = string.byte(str, currentIndex)
    if char > 240 then
      currentIndex = currentIndex + 4
    elseif char > 225 then
      currentIndex = currentIndex + 3
    elseif char > 192 then
      currentIndex = currentIndex + 2
    else 
      currentIndex = currentIndex + 1
    end
    numChars = numChars -1
  end
  return str:sub(start, currentIndex - 1)
end

function BOLTButton:SetNamePlate()
	local r, g, b
	local displayName
	
	self.frameBtn.texture:SetAlpha( 1 )		-- light up the texture
	
	if self.name == "JOTP" then return end
	
	if RAID_CLASS_COLORS[self.targetClass] then
		r = RAID_CLASS_COLORS[self.targetClass].r
		g = RAID_CLASS_COLORS[self.targetClass].g
		b = RAID_CLASS_COLORS[self.targetClass].b
	else
		-- default pale green for unknown classes
		r = 0.75
		g = 1.0
		b = 0.75
	end
	displayName = utf8sub( self.targetName, 1, 5 )	-- TRUNC at 5 chars length
	self.frameBtn.namePlate:SetText( displayName )
	self.frameBtn.namePlate:SetTextColor( r, g, b, 1 )
end

function BOLTButton:ShowMacroTarget()
	local displayName
	
	if self.name == "JOTP" then return end
	
	-- show name plate if the alphalevel of the frame button is not zero
	if self.frameBtn.texture:GetAlpha() ~= 0 then
		if ( self.currentDuration == nil ) or ( self.currentDuration <= 0 ) then
			displayName = utf8sub( self:GetSv( "macroTarget" ), 1, 5 )	-- TRUNC at 5 chars length
			self.frameBtn.namePlate:SetText( displayName )
			self.frameBtn.namePlate:SetTextColor( 0.5, 0.5, 0.5, 1 )	-- default gray color for not active name plates
		end
	end
end

function BOLTButton:SetCounter()
	local red, green, power
	-- local f
	
	-- build the color ramp from green to red, blue value is always 0
	power = self.currentDuration / self.duration
	
	-- if ( power >= 0.5 ) and ( power <= 1 ) then
		-- green = 1.0
		-- red = 1.0 - 2 * ( power - 0.5 )
	-- else
		-- if ( power >= 0 ) and ( power < 0.5 ) then 
			-- red = 1.0
			-- green = 2 * power				
		-- end
	-- end
	
	-- this piece of code build a better color ramp than previous
	if power <= 0.5 then
        green = 4.0 * (power - power ^ 2)
        red = 1.0
    else
        green = 1.0 
        red = 1.0 - 2.0 * ( power - 0.5 )
    end
	
	-- another nice ramp
	-- f = 6.75 * (power - power ^ 1.5)                   
    -- if power < 0.44 then
		-- red = 1
        -- green = f
    -- else
        -- green = 1
        -- red = f
    -- end
		
	self.frameBtn.durationText:SetText( self.currentDuration )	
	self.frameBtn.durationText:SetTextColor( red, green, 0, 1 )
end

function BOLTButton:Reset()
	self.frameBtn.durationText:SetAlpha( 0 )	-- hide the counter
	self.frameBtn.durationText:SetText( "" )	-- reset the counter
	if ReadSv( "generalOptions.alwaysShowNamePlates" ) then
		self:ShowMacroTarget()
	else
		self.frameBtn.namePlate:SetAlpha( 0 )		-- hide the name plate
		self.frameBtn.namePlate:SetText( "" )		-- reset the name plate
	end
end

-- dim the button but only if not inside a duration count
function BOLTButton:Dim()
	if ( self.currentDuration == nil ) or ( self.currentDuration <= 0 ) then
		self.frameBtn.texture:SetAlpha( ReadSv( "generalOptions.dimAlpha" ) )
	
		if ReadSv( "generalOptions.alwaysShowNamePlates" ) then
			self.frameBtn.namePlate:SetAlpha( 1 )
		else
			self.frameBtn.namePlate:SetAlpha( 0 )
		end
	end
end

function BOLTButton:Hide()
	if ( self.currentDuration == nil ) or ( self.currentDuration <= 0 ) then
		self.frameBtn.texture:SetAlpha( 0 )	
		self.frameBtn.namePlate:SetAlpha( 0 )	
	end
end

function BOLTButton:RemoveRoundCorners()
	local k = 0.08
	
	self.frameBtn.texture:SetTexCoord( 0 + k, 1 - k, 0 + k, 1 - k )
end

function BOLTButton:SetRoundCorners()
	self.frameBtn.texture:SetTexCoord( 0, 1, 0, 1 )
end

function BOLTButton:GetSv( svPath )
	return ReadSv( self.name .. "_" .. svPath )	
end

function BOLTButton:SetSv( svPath, svValue )
	WriteSv( self.name .. "_" .. svPath, svValue )
end

function BOLTButton:ToggleBoolSv( svPath )
	return ToggleBoolSv( self.name .. "_" .. svPath )
end